/*
 * A memory pool allocating a region at a time from the system.
 *
 * Copyright (C) 2021, 2022, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/mem_alloc.h>
#include <l4/re/c/rm.h>

#include "memory_incremental.h"

#include <ipc/mem_ipc.h>

#include <stdlib.h>



/* Initialise the memory pool with an optional 'limit' in pages. */

MemoryIncremental::MemoryIncremental(unsigned int limit, offset_t region_size,
                                     l4_cap_idx_t dma)
: _limit(limit), _region_size(region_size), _dma(dma)
{
    _limited = true;
}

MemoryIncremental::MemoryIncremental(l4_cap_idx_t dma)
: _region_size(PAGE_SIZE), _dma(dma)
{
    _limited = false;
}

/* Allocate a block of the given 'size'. */

Region *MemoryIncremental::allocate(offset_t size)
{
    /* Attempt to allocate aligned memory. */

    void *current = NULL;
    l4re_ds_t ds;

    /* Make the size appropriate for the invocation. */

    size = round_multiple(size, PAGE_SIZE);

    /* Use allocation permitting executable mapping of the memory.
       NOTE: Here, it might be beneficial to employ an allocator that obtains
             dataspaces and provides multiple blocks from each dataspace. */

    const l4_size_t attach_flags = L4RE_RM_F_SEARCH_ADDR | L4RE_RM_F_RWX;
    const l4_size_t alloc_flags = l4_is_valid_cap(_dma) ?
                                  L4RE_MA_CONTINUOUS | L4RE_MA_PINNED : 0;

    long err = ipc_new_dataspace(size, alloc_flags, page_order(size), &ds);

    if (err)
        return NULL;

    err = ipc_attach_dataspace_align(ds, size, attach_flags, page_order(size),
                                     (void **) &current);

    if (err)
        return NULL;

    /* Obtain physical addresses if a DMA space capability is available. */

    l4re_dma_space_dma_addr_t paddr = 0;
    l4_size_t size_out = size;

    if (l4_is_valid_cap(_dma))
    {
        long err = l4re_dma_space_map(_dma, ds | L4_CAP_FPAGE_RW, 0, &size_out,
                                      0, L4RE_DMA_SPACE_BIDIRECTIONAL, &paddr);

        if (err)
            return NULL;

        if (size_out < size)
            return NULL;
    }

    return new Region((offset_t) current, (offset_t) current + size, paddr);
}

/* Deallocate the given region. */

void MemoryIncremental::deallocate(Region *region)
{
    ipc_detach_dataspace((void *) region->start);
    delete region;
}

/* Allocate a new region of the given 'size' rounded to the nearest page. */

Region *MemoryIncremental::region(offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    offset_t rounded = round(size, PAGE_SIZE);
    offset_t pages = rounded / PAGE_SIZE;

    /* Check for sufficient pages. */

    if (!_limited || (_limit >= pages))
    {
        /* Attempt to allocate aligned memory. */

        Region *region = allocate(rounded);

        if (region == NULL)
            return NULL;

        if (_limited)
            _limit -= pages;

        return region;
    }

    /* Return no region without sufficient pages. */

    else
        return NULL;
}

/* Allocate a new region having the default region size rounded to the nearest
   page. */

Region *MemoryIncremental::region()
{
    return region(_region_size);
}

/* Release the allocated 'region'. */

void MemoryIncremental::release(Region *region)
{
    std::lock_guard<std::mutex> guard(_lock);

    offset_t rounded = round(region->size(), PAGE_SIZE);
    offset_t pages = rounded / PAGE_SIZE;

    if (_limited)
        _limit += pages;

    deallocate(region);
}

// vim: tabstop=4 expandtab shiftwidth=4
